/*
*
* Created: 15.04.2018 11:30:56
* Author : Yorck Thiele
* ----------------------------------------------------------------------------
**************** GALmate V1.1 ***************************

ATMEGA16 @ 14,7456MHz

V1.1: GAL18V10 pin assignments corrected for RA0..2
* ----------------------------------------------------------------------------
*
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#define PWR_ON_MSG "GALmate_V1.0"
#define VCC_EN PB0 //active low
#define P2_VPP PB1
#define DUT10_GND PB2
#define LED PB3

#define PWM2 PD4
#define PWM1 PD5


#define GAL16V8 0
#define GAL16V8AZ 0x1A
#define ATF16V8B 0x80
#define ATF16V8C 0x81
#define GAL20V8 0x20
#define GAL20V8AZ 0x3A
#define ATF20V8B 0x82
#define GAL6001 0x40
#define GAL6001B 0x41
#define GAL6002B 0x44
#define GAL22V10 0x48
#define GAL22V10AZ 0x49
#define ATF22V10B 0x83
#define ATF22V10C 0x84
#define GAL18V10 0x50
#define GAL18V10B 0x51
#define GAL26CV12 0x58
#define GAL26CV12B 0x59
#define GAL26V12C 0x5D
#define GAL20RA10 0x60
#define GAL20RA10B 0x61
#define GAL20XV10 0x65
#define GAL20XV10B 0x66

#define ATMEGA_PORTA 0
#define ATMEGA_PORTB 1
#define ATMEGA_PORTC 2
#define ATMEGA_PORTD 3

//signal name
#define RA0 0
#define RA1 1
#define RA2 2
#define RA3 3
#define RA4 4
#define RA5 5
#define P_V 6
#define SDIN 7
#define SDOUT 8
#define SCLK 9
#define STROBE 10
#define SDOUT_RD 11 //read back SDOUT


//DUT socket is DIL24
#define DUT23 (1<<PA0)
#define DUT23_PORT ATMEGA_PORTA
#define DUT22 (1<<PA1)
#define DUT22_PORT ATMEGA_PORTA
#define DUT21 (1<<PA2)
#define DUT21_PORT ATMEGA_PORTA
#define DUT3 (1<<PA3)
#define DUT3_PORT ATMEGA_PORTA
#define DUT4 (1<<PA4)
#define DUT4_PORT ATMEGA_PORTA
#define DUT5 (1<<PA5)
#define DUT5_PORT ATMEGA_PORTA
#define DUT6 (1<<PA6)
#define DUT6_PORT ATMEGA_PORTA
#define DUT7 (1<<PA7)
#define DUT7_PORT ATMEGA_PORTA

#define DUT1 (1<<PB4)
#define DUT1_PORT ATMEGA_PORTB
#define DUT2 (1<<PB6)
#define DUT2_PORT ATMEGA_PORTB

#define MUX_S0 (1<<PC0)
//#define DUT13_PORT ATMEGA_PORTC
#define MUX_S1 (1<<PC1)
//#define DUT14_PORT ATMEGA_PORTC
#define MUX_S2 (1<<PC2)
//#define DUT11_PORT ATMEGA_PORTC
#define DUT13_16 (1<<PC3) // SDOUT & STB
#define DUT13_16_PORT ATMEGA_PORTC
#define DUT11 (1<<PC4)
#define DUT11_PORT ATMEGA_PORTC
#define DUT10 (1<<PC5)
#define DUT10_PORT ATMEGA_PORTC
#define DUT9 (1<<PC6)
#define DUT9_PORT ATMEGA_PORTC
#define DUT8 (1<<PC7)
#define DUT8_PORT ATMEGA_PORTC


#define DUT16_RD (1<<PD2)
#define DUT16_RD_PORT ATMEGA_PORTD
#define DUT15_RD (1<<PD3)
#define DUT15_RD_PORT ATMEGA_PORTD
#define DUT14_RD (1<<PD6)
#define DUT14_RD_PORT ATMEGA_PORTD
#define DUT9_RD (1<<PD7)
#define DUT9_RD_PORT ATMEGA_PORTD

#define NOT_TIMEOUT_T0 !(TIFR & (1<<TOV0)) //==0
#define NOT_TIMEOUT_T1 !(TIFR & (1<<TOV1)) //==0
#define TIMEOUT_T1 (TIFR & (1<<TOV1)) //!=0

//#define SPI_BYTE_RDY (SPSR & (1<<SPIF))
//#define NOT_SPI_BYTE_RDY !(SPSR & (1<<SPIF))



volatile uint8_t Signal_Pin[22];//see signal names
volatile uint8_t Signal_Port[22];
volatile uint8_t Mux_Chan;
volatile uint8_t rxbuffer[260];
//volatile uint8_t tx_rx_buffer[340];
volatile uint8_t sectorbuffer[256];
//volatile uint16_t tx_rx_buffer_ptr;
//volatile uint8_t rxbuff;
volatile uint16_t bytes2pc;

/*
volatile uint8_t sec_buff_ptr;
volatile uint8_t rxbuffindex;

volatile uint8_t frameerror;
*/
volatile uint8_t status;
volatile uint8_t pvalid;
volatile uint8_t pid;
volatile uint8_t plen;
volatile uint8_t gal_id;
volatile uint8_t gal_pins;
//===================================================================================
void
set_GAL_pinout(void)
{
	switch(gal_id)
	{
		case GAL16V8:
		case GAL16V8AZ:
		case ATF16V8B:
		case ATF16V8C:
		Signal_Pin[RA0]=DUT22;
		Signal_Port[RA0]=DUT22_PORT;
		Signal_Pin[RA1]=DUT3;
		Signal_Port[RA1]=DUT3_PORT;
		Signal_Pin[RA2]=DUT4;
		Signal_Port[RA2]=DUT4_PORT;
		Signal_Pin[RA3]=DUT5;
		Signal_Port[RA3]=DUT5_PORT;
		Signal_Pin[RA4]=DUT6;
		Signal_Port[RA4]=DUT6_PORT;
		Signal_Pin[RA5]=DUT7;
		Signal_Port[RA5]=DUT7_PORT;
		Signal_Pin[P_V]=DUT23;
		Signal_Port[P_V]=DUT23_PORT;
		Signal_Pin[SDIN]=DUT9;
		Signal_Port[SDIN]=DUT9_PORT;
		Signal_Pin[SDOUT]=DUT13_16;
		Signal_Port[SDOUT]=DUT13_16_PORT;
		Signal_Pin[SCLK]=DUT8;
		Signal_Port[SCLK]=DUT8_PORT;
		Signal_Pin[STROBE]=DUT13_16;
		Signal_Port[STROBE]=DUT13_16_PORT;
		Signal_Pin[SDOUT_RD]=DUT16_RD;
		Signal_Port[SDOUT_RD]=DUT16_RD_PORT;
		break;

		case GAL20V8:
		case GAL20V8AZ:
		case ATF20V8B:
		Signal_Pin[RA0]=DUT21;
		Signal_Port[RA0]=DUT21_PORT;
		Signal_Pin[RA1]=DUT3;
		Signal_Port[RA1]=DUT3_PORT;
		Signal_Pin[RA2]=DUT4;
		Signal_Port[RA2]=DUT4_PORT;
		Signal_Pin[RA3]=DUT5;
		Signal_Port[RA3]=DUT5_PORT;
		Signal_Pin[RA4]=DUT8;
		Signal_Port[RA4]=DUT8_PORT;
		Signal_Pin[RA5]=DUT9;
		Signal_Port[RA5]=DUT9_PORT;
		Signal_Pin[P_V]=DUT22;
		Signal_Port[P_V]=DUT22_PORT;
		Signal_Pin[SDIN]=DUT11;
		Signal_Port[SDIN]=DUT11_PORT;
		Signal_Pin[SDOUT]=DUT13_16;
		Signal_Port[SDOUT]=DUT13_16_PORT;
		Signal_Pin[SCLK]=DUT10;
		Signal_Port[SCLK]=DUT10_PORT;
		Signal_Pin[STROBE]=DUT13_16;
		Signal_Port[STROBE]=DUT13_16_PORT;
		Signal_Pin[SDOUT_RD]=DUT15_RD;
		Signal_Port[SDOUT_RD]=DUT15_RD_PORT;
		break;

		case GAL22V10:
		case GAL22V10AZ:
		case ATF22V10B:
		case ATF22V10C:
		case GAL20RA10:
		case GAL20RA10B:
		case GAL20XV10:
		case GAL20XV10B:
		case GAL26CV12:
		case GAL26CV12B:
		case GAL26V12C:
		Signal_Pin[RA0]=DUT4;
		Signal_Port[RA0]=DUT4_PORT;
		Signal_Pin[RA1]=DUT5;
		Signal_Port[RA1]=DUT5_PORT;
		Signal_Pin[RA2]=DUT6;
		Signal_Port[RA2]=DUT6_PORT;
		Signal_Pin[RA3]=DUT7;
		Signal_Port[RA3]=DUT7_PORT;
		Signal_Pin[RA4]=DUT8;
		Signal_Port[RA4]=DUT8_PORT;
		Signal_Pin[RA5]=DUT9;
		Signal_Port[RA5]=DUT9_PORT;
		Signal_Pin[P_V]=DUT3;
		Signal_Port[P_V]=DUT3_PORT;
		Signal_Pin[SDIN]=DUT11;
		Signal_Port[SDIN]=DUT11_PORT;
		Signal_Pin[SDOUT]=DUT13_16;
		Signal_Port[SDOUT]=DUT13_16_PORT;
		Signal_Pin[SCLK]=DUT10;
		Signal_Port[SCLK]=DUT10_PORT;
		Signal_Pin[STROBE]=DUT13_16;
		Signal_Port[STROBE]=DUT13_16_PORT;
		Signal_Pin[SDOUT_RD]=DUT14_RD;
		Signal_Port[SDOUT_RD]=DUT14_RD_PORT;
		break;
		
		case GAL18V10:
		case GAL18V10B:

		Signal_Pin[RA0]=DUT22;
		Signal_Port[RA0]=DUT22_PORT;
		Signal_Pin[RA1]=DUT21;
		Signal_Port[RA1]=DUT21_PORT;
		Signal_Pin[RA2]=DUT13_16;
		Signal_Port[RA2]=DUT13_16_PORT;
		Signal_Pin[RA3]=DUT3;
		Signal_Port[RA3]=DUT3_PORT;
		Signal_Pin[RA4]=DUT4;
		Signal_Port[RA4]=DUT4_PORT;
		Signal_Pin[RA5]=DUT5;
		Signal_Port[RA5]=DUT5_PORT;
		Signal_Pin[P_V]=DUT23;
		Signal_Port[P_V]=DUT23_PORT;
		Signal_Pin[SDIN]=DUT7;
		Signal_Port[SDIN]=DUT7_PORT;
		Signal_Pin[SDOUT]=DUT9;
		Signal_Port[SDOUT]=DUT9_PORT;
		Signal_Pin[SCLK]=DUT6;
		Signal_Port[SCLK]=DUT6_PORT;
		Signal_Pin[STROBE]=DUT8;
		Signal_Port[STROBE]=DUT8_PORT;
		Signal_Pin[SDOUT_RD]=DUT9_RD;
		Signal_Port[SDOUT_RD]=DUT9_RD_PORT;
		break;
		
		case GAL6001:
		case GAL6001B:
		case GAL6002B:
		Signal_Pin[RA0]=DUT4;
		Signal_Port[RA0]=DUT4_PORT;
		Signal_Pin[RA1]=DUT5;
		Signal_Port[RA1]=DUT5_PORT;
		Signal_Pin[RA2]=DUT6;
		Signal_Port[RA2]=DUT6_PORT;
		Signal_Pin[RA3]=DUT7;
		Signal_Port[RA3]=DUT7_PORT;
		Signal_Pin[RA4]=DUT8;
		Signal_Port[RA4]=DUT8_PORT;
		Signal_Pin[RA5]=DUT9;
		Signal_Port[RA5]=DUT9_PORT;
		Signal_Pin[P_V]=DUT23;
		Signal_Port[P_V]=DUT23_PORT;
		Signal_Pin[SDIN]=DUT11;
		Signal_Port[SDIN]=DUT11_PORT;
		Signal_Pin[SDOUT]=DUT13_16;
		Signal_Port[SDOUT]=DUT13_16_PORT;
		Signal_Pin[SCLK]=DUT10;
		Signal_Port[SCLK]=DUT10_PORT;
		Signal_Pin[STROBE]=DUT13_16;
		Signal_Port[STROBE]=DUT13_16_PORT;
		Signal_Pin[SDOUT_RD]=DUT14_RD;
		Signal_Port[SDOUT_RD]=DUT14_RD_PORT;
		break;
	}
}

//===================================================================================
uint8_t
get_GAL_SDOUT(void)
{

	if((PIND & Signal_Pin[SDOUT_RD])==0) return 0; else return 1;

}

//===================================================================================
void
set_GAL_signal(uint8_t signal_name, uint8_t signal_val)
{
	switch(Signal_Port[signal_name])
	{
		case ATMEGA_PORTA:
		if(signal_val==0) PORTA &= ~Signal_Pin[signal_name]; else PORTA |= Signal_Pin[signal_name];
		break;
		case ATMEGA_PORTB:
		if(signal_val==0) PORTB &= ~Signal_Pin[signal_name]; else PORTB |= Signal_Pin[signal_name];
		break;
		case ATMEGA_PORTC:
		if(signal_val==0) PORTC &= ~Signal_Pin[signal_name]; else PORTC |= Signal_Pin[signal_name];
		break;
		case ATMEGA_PORTD:
		if(signal_val==0) PORTD &= ~Signal_Pin[signal_name]; else PORTD |= Signal_Pin[signal_name];
		break;
	}
}
//===================================================================================
void
set_GAL_SDIN(uint8_t logic_level)
{
	if (logic_level ==0) set_GAL_signal(SDIN,0); else set_GAL_signal(SDIN,1);
}
//===================================================================================
void
set_GAL_row(uint8_t row)
{
	set_GAL_signal(RA0,row & 1);
	set_GAL_signal(RA1,row & 2);
	set_GAL_signal(RA2,row & 4);
	set_GAL_signal(RA3,row & 8);
	set_GAL_signal(RA4,row & 16);
	set_GAL_signal(RA5,row & 32);
}
//===================================================================================
ISR(USART_RXC_vect)
{
	
	static uint16_t sec_buff_ptr;
	static uint8_t rxbuffindex;
	static uint8_t frameerror;
	static uint8_t pages;

	uint8_t serbyte;


	if ((UCSRA & (1<<FE)) == 0)
	{
		serbyte = UDR;
		if (frameerror!=0) {rxbuffindex=0;}
		switch(rxbuffindex)
		{
			case 0:
			pid=serbyte;
			sec_buff_ptr=0;
			plen=0;
			break;
			case 1:
			plen=serbyte;
			break;
			case 2:
			pages=serbyte;
			break;
			case 3:
			if(plen==0)
			{
				sectorbuffer[sec_buff_ptr++]=serbyte;
				if((sec_buff_ptr/256)==pages) pvalid=1; // pages x 256 bytes received?
			}
			break;
		}
		if((plen!=0) || (rxbuffindex<3))
		{
			rxbuffer[rxbuffindex++]=serbyte;
			if(rxbuffindex==plen) pvalid=1;
		}
		frameerror=0;
	}
	else
	{
		serbyte = UDR;
		if (frameerror!=0) {wdt_enable(WDTO_15MS);for(;;);} // 2 x FRAME ERROR --> RESET ATMEGA
		frameerror=1;
	}
}
//===================================================================================
void
delay_n(uint16_t n)
{
	//n x 0.54us
	//32ms max
	TCNT0=255-n;
	TIFR=(1<<TOV0);
	while (NOT_TIMEOUT_T0);
}
//===================================================================================
void
delay_us(uint8_t usec)
{
	//n x 1.085us
	//138us max
	TCNT0=255-usec*2+2;
	TIFR=(1<<TOV0);
	while (NOT_TIMEOUT_T0);
}
//===================================================================================
void
delay_ms(uint16_t msec)
{
	while (msec != 0)
	{
		delay_us(125);
		delay_us(125);
		delay_us(125);
		delay_us(125);
		delay_us(125);
		delay_us(125);
		delay_us(125);
		delay_us(125);
		msec--;
	}
}
//===================================================================================
void
set_timeout(uint16_t usec)
{
	//n x 1.085us
	//32ms max
	TCNT1=65535-usec*2;
	TIFR=(1<<TOV1);
}
//===================================================================================
void
send_byte_to_PC(uint8_t byte)
{
	while (!(UCSRA & (1<<UDRE)));
	UDR=byte;
	bytes2pc++;
}
//===================================================================================
void
frame_termination(uint16_t returnbytes)
{
	uint16_t payload=bytes2pc;

	while(bytes2pc < returnbytes)
	{
		send_byte_to_PC(0);
	}
	send_byte_to_PC(payload & 255);
	send_byte_to_PC(payload / 256);
	send_byte_to_PC(status);
	send_byte_to_PC(pid);
}
//===================================================================================
void
send_string_RS232(char *txtptr)
{
	uint8_t z;

	for (z=0;txtptr[z]!=0;z++)
	{
		send_byte_to_PC(txtptr[z]); // send byte via USART
	}
}
//===================================================================================
void
init(void)
{
	cli();
	wdt_reset();
	MCUCSR &= ~(1<<WDRF);
	wdt_disable();


	DDRA=255;
	DDRB=254;
	DDRC=255;
	DDRD = (1 << PWM1) | (1 << PWM2);


	//IECPORT_D = (1 << TRIGGER);
	//Timer0
	TCCR0=(1 << CS01);// CLK (68ns * 8 = 544ns) for Timer0

	//Timer1
	TCCR1A=(1 << COM1A1) | (1 << COM1B1) | (1 << WGM10);
	TCCR1B=(1 << WGM12) | (1 << CS10);// CLK/1 (68ns) for Timer1; fast PWM 8-bit
	OCR1A=255;
	OCR1B=178;

	//UART
	UBRRH=0;
	UBRRL=7; //115200 Baud @ 14,7456MHz
	UCSRB=(1<<RXEN) | (1<<TXEN) | (1<<RXCIE);
	UCSRC= (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0); // no parity, 1 stop bit, 8 data bits


	delay_ms(1);
	send_string_RS232(PWR_ON_MSG);
	sei(); //enable RX interrupt

}
//===================================================================================
void
set_EDIT_mode(uint8_t on_off)
{
	if(on_off!=0)
	{
		PORTB &= ~(1<<P2_VPP); //disable EDIT
		delay_ms(5);
		DDRB &= ~(1<<VCC_EN); //disable VCC
		PORTA=0;
		PORTB=0;
		PORTC=0;
		delay_ms(1);


		switch(gal_pins)
		{
			case 20:
			PORTB |= (1<<DUT10_GND); //enable GND
			break;
			default:
			PORTB &= ~(1<<DUT10_GND); //disable GND
		}
		
		DDRB |= (1<<VCC_EN); //enable VCC
		PORTB |= (1<<LED);
		//check for pull-up or bus keeper
		DDRA=0;
		DDRC &= ~(1<<PC1);
		delay_ms(1);
		if((PINA & DUT7) !=0)
		{
			send_byte_to_PC(1); //pull-up detected
		}
		else
		{
			DDRC |= (1<<PC1);
			PORTC |= (1<<PC1); //PC1 via 1MOhm to DUT pin7
			delay_ms(1);
			if((PINA & DUT7) ==0) send_byte_to_PC(2); //bus keeper detected
			DDRC &= ~(1<<PC1);
			PORTC &= ~(1<<PC1);
		}
		DDRA=255;

		set_GAL_signal(STROBE,1);
		set_GAL_signal(SDOUT,1);
		//set_GAL_signal(P_V,1); //7.2.2019
		delay_ms(5);
		PORTB |= (1<<P2_VPP); //enable EDIT
		delay_ms(1);

	}
	else
	{
		PORTB &= ~(1<<P2_VPP); //disable EDIT
		delay_ms(5);
		set_GAL_row(0);
		set_GAL_signal(SDIN,0);
		set_GAL_signal(SDOUT,0);
		set_GAL_signal(SCLK,0);
		set_GAL_signal(P_V,0);
		set_GAL_signal(STROBE,0);
		set_GAL_signal(SDOUT,0);
		DDRB &= ~(1<<VCC_EN); //disable VCC
		PORTB &= ~(1<<DUT10_GND); //disable GND
		PORTB &= ~(1<<LED);
		
	}
}
//===================================================================================
void
read_GAL_row(uint8_t nbits)
{
	uint8_t z, subadr, subadr_width, subadr_order, adrbits;

	subadr=rxbuffer[4];
	subadr_width=rxbuffer[5];
	subadr_order=rxbuffer[6];
	set_GAL_row(rxbuffer[3]);
	adrbits=rxbuffer[7];

	if(adrbits>0) //address setup by row data
	{
		for (z=0;z<adrbits;z++)
		{
			set_GAL_SDIN(rxbuffer[10+z]);
			if(!((subadr_order!=0) &&  (subadr == 255) && (z==adrbits-1))) //skip last clock (ATF22V10C only)
			{
				delay_us(5);
				set_GAL_signal(SCLK,1);
				delay_us(5);
				set_GAL_signal(SCLK,0);
			}
		}
	}
	else
	{
		if(subadr != 255) //subadr?
		{
			if(rxbuffer[6]==0) //subadr bit order
			{
				for (z=1;z<=subadr_width;z++)//send n subadr bits, LSB first
				{
					set_GAL_SDIN(subadr & 1);
					subadr=subadr>>1;
					delay_us(5);
					set_GAL_signal(SCLK,1);
					delay_us(5);
					set_GAL_signal(SCLK,0);
				}
			}
			else //ATMEL ATF22V10C
			{
				for (z=(1<<(subadr_width-1));z!=0;z>>=1)//send n subadr bits, MSB first
				{
					set_GAL_SDIN(subadr & z);
					if(z>1)
					{
						delay_us(5);
						set_GAL_signal(SCLK,1);
						delay_us(5);
						set_GAL_signal(SCLK,0);
					}
				}
			}
		}
	}
	set_GAL_signal(P_V,0); //27.2.2019
	delay_us(5);
	set_GAL_signal(STROBE,0);
	delay_us(5);
	set_GAL_signal(STROBE,1);
	set_GAL_SDIN(0); // VERY IMPORTANT!
	delay_us(5);
	if(gal_id==ATF16V8C) set_GAL_signal(P_V,1); //27.2.2019
	for (z=0;z<nbits;z++)
	{
		delay_us(5);
		send_byte_to_PC(get_GAL_SDOUT()); // send one byte for every bit read via USART
		set_GAL_signal(SCLK,1);
		delay_us(5);
		set_GAL_signal(SCLK,0);
	}
	set_GAL_signal(P_V,0); //27.2.2019
	
	
}
//===================================================================================
void
write_GAL_row(void)
{
	uint8_t z, subadr, subadr_width, subadr_order, nbits;


	subadr=rxbuffer[4];
	subadr_width=rxbuffer[5];
	subadr_order=rxbuffer[6];
	set_GAL_row(rxbuffer[3]);
	nbits=rxbuffer[2];
	
	if(gal_id==ATF16V8C)
	{
		set_GAL_signal(P_V,1); //27.2.2019
		for (z=0;z<nbits;z++)
		{
			set_GAL_SDIN(rxbuffer[10+z]);
			delay_us(5);
			set_GAL_signal(SCLK,1);
			if(!(z==nbits-1)) //skip last clock low transition
			{
				delay_us(5);
				set_GAL_signal(SCLK,0);
			}
		}
	}
	else
	{
		for (z=0;z<nbits;z++)
		{
			set_GAL_SDIN(rxbuffer[10+z]);
			if(!((subadr_order!=0) &&  (subadr == 255) && (z==nbits-1))) //skip last clock (ACW row ATF22V10C only)
			{
				delay_us(5);
				set_GAL_signal(SCLK,1);
				delay_us(5);
				set_GAL_signal(SCLK,0);
			}
		}
		
		if(subadr != 255) //subadr?
		{
			if(subadr_order==0) //subadr has regular bit order
			{
				for (z=1;z<=subadr_width;z++)//send n bits, LSB first
				{
					set_GAL_SDIN(subadr & 1);
					subadr=subadr>>1;
					delay_us(5);
					set_GAL_signal(SCLK,1);
					delay_us(5);
					set_GAL_signal(SCLK,0);
				}
				set_GAL_SDIN(0); // VERY IMPORTANT!
			}
			else //ATMEL ATF22V10C has reverse bit order
			{
				for (z=(1<<(subadr_width-1));z!=0;z>>=1)//send n bits, MSB first
				{
					set_GAL_SDIN(subadr & z);
					if(z>1)
					{
						delay_us(5);
						set_GAL_signal(SCLK,1);
						delay_us(5);
						set_GAL_signal(SCLK,0);
					}
				}
			}
		}
		if (rxbuffer[8]==0x80) {rxbuffer[8]=0; set_GAL_SDIN(0);} // VERY IMPORTANT!

	}
	delay_us(5);
	set_GAL_signal(P_V,1);
	set_GAL_signal(STROBE,0);
	delay_ms(rxbuffer[8] * 256 + rxbuffer[7]); // max 1s
	set_GAL_signal(STROBE,1);
	set_GAL_signal(P_V,0);
	set_GAL_signal(SCLK,0);


}
//===================================================================================
void
detect_device(void)
{
	uint8_t z;
	
	if(gal_id==ATF16V8C) set_GAL_signal(P_V,1); //27.2.2019
	set_GAL_signal(SDIN,1);
	for (z=1;z!=0;z++)
	{
		delay_us(5);
		set_GAL_signal(SCLK,1);
		delay_us(5);
		set_GAL_signal(SCLK,0);
	}

	set_GAL_signal(SDIN,0);
	for (z=1;z!=0;z++)
	{
		delay_us(5);
		set_GAL_signal(SCLK,1);
		delay_us(5);
		set_GAL_signal(SCLK,0);
		if(get_GAL_SDOUT()==0) break;
	}
	send_byte_to_PC(z); //byte 0
	
	if(z>0)
	{
		set_GAL_signal(SDIN,1);
		for (z=1;z!=0;z++)
		{
			delay_us(5);
			set_GAL_signal(SCLK,1);
			delay_us(5);
			set_GAL_signal(SCLK,0);
			if(get_GAL_SDOUT()==1) break;
		}
	}

	send_byte_to_PC(z); //byte 1
	set_GAL_signal(P_V,0);
	set_GAL_signal(SDIN,0);


}
//===================================================================================
void
power_up(void)
{
	uint8_t n;
	
	if(rxbuffer[2]>100) //VCC >3.3V?
	{
		if(rxbuffer[6]==0)
		{
			OCR1B=rxbuffer[2]; //VCC
			OCR1A=rxbuffer[3]; //VEDIT
			gal_id=rxbuffer[4];
			gal_pins=rxbuffer[5];
			set_GAL_pinout();
			set_EDIT_mode(1);
		}
		else //ramp VEDIT
		//avoid high dv/dt (especially for ST GALs)
		{
			OCR1B=rxbuffer[2]; //VCC
			OCR1A=rxbuffer[7]; //VPP_READ
			gal_id=rxbuffer[4];
			gal_pins=rxbuffer[5];
			set_GAL_pinout();
			set_EDIT_mode(1);
			if(rxbuffer[7]!=rxbuffer[3])
			{
				for (n=0;n<8;n++)
				{
					OCR1A=OCR1A+(rxbuffer[3]-rxbuffer[7])/8;
					delay_ms(5);
				}
			}
			OCR1A=rxbuffer[3];//VPP
		}
	}
}
//===================================================================================
int
main (void)
{
	uint16_t returnbytes;
	
	init();


	for (;;)
	{

		if(pvalid!=0)
		{
			returnbytes=16;
			bytes2pc=0;

			switch(pid)
			{
				case 1: //power-up
				power_up();
				break;
				
				case 2: //shut-down
				set_EDIT_mode(0);
				break;
				
				case 3: //detect GAL (i.e its shift register)
				detect_device();
				set_EDIT_mode(0);
				break;
				
				case 5: //read row
				returnbytes=rxbuffer[2];
				read_GAL_row(returnbytes);
				break;
				
				case 7: //write row
				if(rxbuffer[9]==0)
				{
					write_GAL_row();
				}
				else
				{
					while(rxbuffer[9]!=0)
					{
						write_GAL_row();
						rxbuffer[9]--;
						if(rxbuffer[5]==0) rxbuffer[3]++; else rxbuffer[4]++; //handle subadr
					}
				}
				break;

			}
			frame_termination(returnbytes);
			pvalid=0;
		}
	}
}

